ALERT_DIALOG_DESCRIPTION = '//android.widget.TextView[@resource-id="com.snapchat.android:id/alert_dialog_description"]'
CAMERA_CAPTURE = '//android.widget.FrameLayout[@content-desc="Camera Capture"]'
CAMERA_PAGE = '//android.widget.FrameLayout[@resource-id="com.snapchat.android:id/camera_page"]'
CAPTION_EDIT = '//android.widget.EditText[@resource-id="com.snapchat.android:id/caption_edit_text_view"]'
CHAT_INPUT = '//android.widget.EditText[@resource-id="com.snapchat.android:id/chat_input_text_field"]'
CHAT_CONVERSATION = ('//android.view.View['
                     '@resource-id="com.snapchat.android:id/ff_item"'
                     ' and .//javaClass[@text="{conversation}"]]')
CHAT_TAB = '//android.view.ViewGroup[@content-desc="Chat"]'
CONVERSATION_TITLE = '//android.widget.TextView[@resource-id="com.snapchat.android:id/conversation_title_text_view"]'
CONVERSATION_TITLE_TEXT = '//android.widget.TextView[@resource-id="com.snapchat.android:id/conversation_title_text_view" and contains(lower-case(@text), "{conversation}")]'
DISCARD_BUTTON = '//android.widget.LinearLayout[@resource-id="com.snapchat.android:id/dialog_content"]/android.view.View'
FEED_NEW_CHAT = '//android.widget.FrameLayout[@resource-id="com.snapchat.android:id/feed_new_chat"]'
FULL_SCREEN_SURFACE_VIEW = '//android.view.SurfaceView[@resource-id="com.snapchat.android:id/full_screen_surface_view"]'
RECIPIENTS_TO_ADD = ('//androidx.recyclerview.widget.RecyclerView[@resource-id="com.snapchat.android:id/send_to_recycler_view"]'
                  '//android.view.View[count(.//javaClass)=1]//javaClass[@text="{recipient}"]')
MY_STORY = '//javaClass[@text="My Story · Friends Only"]/..'
NEW_STORY = '//android.view.View[@content-desc="New Story Button"]'
SEND_BUTTON = '//android.view.View[@content-desc="Send"]'
SENT_TO = '//android.view.ViewGroup[@resource-id="com.snapchat.android:id/sent_to_button_label_mode_view"]'
TOGGLE_CAMERA = '(//android.widget.ImageView[@resource-id="com.snapchat.android:id/camera_mode_icon_image_view"])[1]'
